#ifndef MSP_DATAFILE_LOADER_H_
#define MSP_DATAFILE_LOADER_H_

#include <type_traits>
#include <map>
#include <msp/io/file.h>
#include "loaderaction.h"
#include "meta.h"
#include "parser.h"
#include "statement.h"

namespace Msp {
namespace DataFile {

/**
Base class for data loaders.  This class only provides core functionality.
You'll almost certainly want to use one of the BasicLoader classes instead.

Under normal circumstances, a class capable of being loaded should have a
nested typed called Loader which resolves to a descendant of this class.  If
another structure is used, the loader object must be constructed manually.

A loader class should execute one or more calls to the various add() functions
to register actions with expected keywords.  Currently possible actions are
calling a function of the loader, storing values in member variables of an
object and ignoring the statement.  If a unexpected keyword is encountered, an
exception is thrown and the loading is aborted.

A sub-object can be loaded with one of the load_sub functions.

When loading has finished successfully, the virtual function finish() is
called.  Any post-processing of the data should be placed here and not in the
destructor.

See also classes ObjectLoader and CollectionObjectLoader in objectloader.h.
*/
class Loader: private NonCopyable
{
protected:
	class ActionMap: public std::map<StatementKey, LoaderAction *>, private NonCopyable
	{
	public:
		~ActionMap();
	};

private:
	ActionMap local_actions;
	ActionMap *actions = nullptr;
	Parser *cur_parser = nullptr;
	unsigned cur_level = 0;
	const Statement *cur_st = nullptr;
	bool sub_loaded = false;
	bool direct = false;
	std::vector<Loader *> aux_loaders;
protected:
	bool check_sub_loads = false;

	Loader() = default;
public:
	virtual ~Loader() = default;

	/** Loads statements from a parser. */
	void load(Parser &p);

private:
	/** Loads data from a statement. */
	void load(const Statement &st);

	/** Loads statemsnts from a parser, feeding them directly to actions. */
	void load_direct(Parser &, unsigned);

	/** Processes a single statement */
	void load_statement(const Statement &st);

protected:
	/** Loads a sub-object from the statement being processed.  The Loader class
	of the sub-object is automatically used. */
	template<typename S>
	void load_sub(S &s)
	{
		typename S::Loader ldr(s);
		load_sub_with(ldr);
	}

	/** Loads a sub-object from the statement being processed with an extra
	parameter for the Loader.  The Loader class of the sub-object is
	automatically used. */
	template<typename S, typename... Args>
	void load_sub(S &s, Args &&... args)
	{
		typename S::Loader ldr(s, std::forward<Args>(args)...);
		load_sub_with(ldr);
	}

	/** Processes the current statement's substatements with another Loader. */
	void load_sub_with(Loader &);

	/** Sets the actions to be used when loading.  If the map is empty,
	init_actions will be called. */
	void set_actions(ActionMap &);
	virtual void init_actions() { }

	/** Adds a keyword that is loaded by calling a function. */
	template<typename L>
	void add(const std::string &k, void (L::*func)())
	{ add(k, new LoaderFunc0<L>(func)); }

	template<typename L, typename A0>
	void add(const std::string &k, void (L::*func)(A0))
	{ add(k, new LoaderFunc1<L, A0>(func)); }

	template<typename L, typename... Args>
	void add(const std::string &k, void (L::*func)(Args...))
	{ add(k, new LoaderFuncN<L, Args...>(func)); }

	/** Adds a keyword that is loaded by calling a function with a bound
	first argument. */
	template<typename L, typename B0, typename... Args>
	void add(const std::string &k, void (L::*func)(B0, Args...), const typename std::remove_reference<B0>::type &b0)
	{ add(k, new LoaderFuncNBound1<L, B0, Args...>(func, b0)); }

	template<typename L, typename B0, typename... Args>
	void add(const std::string &k, void (L::*func)(B0, Args...), B0 &&b0)
	{ add(k, new LoaderFuncNBound1<L, B0, Args...>(func, std::forward<B0>(b0))); }

	/** Adds a keyword that is loaded into a member of the loaded object. */
	template<typename L, typename T0>
	void add(const std::string &k, T0 L::*p0)
	{ add(k, new LoadValue1<L, T0>(p0)); }

	template<typename L, typename T0, typename T1>
	void add(const std::string &k, T0 L::*p0, T1 L::*p1)
	{ add(k, new LoadValue2<L, T0, T1>(p0, p1)); }

	/** Adds a keyword that is recognized but ignored. */
	void add(const std::string &k)
	{ add(k, nullptr); }

private:
	void add(const std::string &, LoaderAction *);

protected:
	void add_auxiliary_loader(Loader &);

private:
	bool has_action(const StatementKey &) const;
	LoaderAction *find_action(const StatementKey &) const;

protected:
	/** Returns the source of the statement being processed.  This can be used
	to implement relative paths in include-like statements.  Note that the
	source may not necessarily be a file. */
	const std::string &get_source() const;

	/** Returns the keyword of the statement being processed.  Can be used to
	implement dynamic keywords. */
	const std::string &get_keyword() const;

	virtual void finish() { }
};


/**
Loads an object from a file.  The object must have a public Loader class.  Any
extra arguments are passed to the Loader constructor.
*/
template<typename T, typename... Args>
void load(T &obj, const std::string &fn, Args &&... args)
{
	IO::BufferedFile in(fn);

	Parser parser(in, fn);
	typename T::Loader loader(obj, std::forward<Args>(args)...);
	loader.load(parser);
}

/**
Loads an object from a file stored in a collection.  The object must have a
public Loader class.  The collection is passed to the Loader constructor,
followed by any extra arguments.
*/
template<typename T, typename... Args>
void load(T &obj, typename T::Loader::Collection &coll, const std::string &fn, Args &&... args)
{
	RefPtr<IO::Seekable> in = coll.open_raw(fn);
	if(!in)
		throw IO::file_not_found(fn);

	Parser parser(*in, fn);
	typename T::Loader loader(obj, coll, std::forward<Args>(args)...);
	loader.load(parser);
}

/**
Loads an object from a file stored in a collection.  The object must havea
public Loader class.  Any extra arguments are passed to the Loader constructor.
*/
template<typename T, typename C, typename... Args>
typename std::enable_if<!NeedsCollection<typename T::Loader>::value>::type load(T &obj, C &coll, const std::string &fn, Args &&... args)
{
	RefPtr<IO::Seekable> in = coll.open_raw(fn);
	if(!in)
		throw IO::file_not_found(fn);

	Parser parser(*in, fn);
	typename T::Loader loader(obj, std::forward<Args>(args)...);
	loader.load(parser);
}

} // namespace DataFile
} // namespace Msp

#endif
