#ifndef MSP_DATAFILE_LOADERACTION_H_
#define MSP_DATAFILE_LOADERACTION_H_

#include <msp/core/meta.h>
#include "argumentstore.h"
#include "statement.h"

namespace Msp {
namespace DataFile {

template<typename T>
std::string create_signature(const std::string &prefix = std::string())
{ return prefix+std::string(1, TypeInfo<T>::signature); }

template<typename T0, typename T1, typename... Tail>
std::string create_signature(const std::string &prefix = std::string())
{ return create_signature<T1, Tail...>(prefix+std::string(1, TypeInfo<T0>::signature)); }

class Loader;

/**
Base class for loader actions.
*/
class LoaderAction
{
protected:
	LoaderAction() = default;
public:
	virtual ~LoaderAction() = default;

	/** Called to process a statement. */
	virtual void execute(Loader &, const Statement &) const = 0;

	virtual void execute(Loader &, const ArgumentStore &) const = 0;

	virtual std::string get_signature() const = 0;
};


/**
Loads a statement by calling a function that takes no arguments.
*/
template<typename L>
class LoaderFunc0: public LoaderAction
{
private:
	typedef void (L::*FuncType)();

	FuncType func;

public:
	LoaderFunc0(FuncType f): func(f) { }

	void execute(Loader &l, const Statement &) const override
	{
		(dynamic_cast<L &>(l).*func)();
	};

	void execute(Loader &l, const ArgumentStore &) const override
	{
		(dynamic_cast<L &>(l).*func)();
	};

	std::string get_signature() const override
	{ return std::string(); }
};


/**
Loads a statement by calling a function that takes one argument.
*/
template<typename L, typename A0>
class LoaderFunc1: public LoaderAction
{
private:
	typedef void (L::*FuncType)(A0);

	FuncType func;

public:
	LoaderFunc1(FuncType f): func(f) { }

	void execute(Loader &l, const Statement &st) const override
	{
		(dynamic_cast<L &>(l).*func)(st.args[0].get<A0>());
	}

	void execute(Loader &l, const ArgumentStore &as) const override
	{
		(dynamic_cast<L &>(l).*func)(as.get<A0>(0));
	}

	std::string get_signature() const override
	{ return std::string(1, TypeInfo<A0>::signature); }
};


/**
Loads a statement by calling a function that takes an array of values.
*/
template<typename L, typename A0>
class LoaderFunc1<L, const std::vector<A0> &>: public LoaderAction
{
private:
	typedef void (L::*FuncType)(const std::vector<A0> &);

	FuncType func;

public:
	LoaderFunc1(FuncType f): func(f) { }

	void execute(Loader &l, const Statement &st) const override
	{
		std::vector<A0> values;
		values.reserve(st.args.size());
		for(const Value &a: st.args)
			values.push_back(a.get<A0>());
		(dynamic_cast<L &>(l).*func)(values);
	}

	void execute(Loader &l, const ArgumentStore &as) const override
	{
		std::vector<A0> values;
		unsigned n_args = as.get_info().key.signature.size();
		values.reserve(n_args);
		for(unsigned i=0; i<n_args; ++i)
			values.push_back(as.get<A0>(i));
		(dynamic_cast<L &>(l).*func)(values);
	}

	std::string get_signature() const override
	{
		std::string result;
		result += TypeInfo<A0>::signature;
		result += '*';
		return result;
	}
};


/**
Loads a statement by calling a function with the statement itself as argument.
*/
template<typename L>
class LoaderFunc1<L, const Statement &>: public LoaderAction
{
private:
	typedef void (L::*FuncType)(const Statement &);

	FuncType func;

public:
	LoaderFunc1(FuncType f): func(f) { }

	void execute(Loader &l, const Statement &st) const override
	{
		(dynamic_cast<L &>(l).*func)(st);
	}

	void execute(Loader &, const ArgumentStore &) const override
	{
		throw std::logic_error("incompatible format");
	}

	std::string get_signature() const override
	{ return "*"; }
};


template<unsigned I, typename... Args>
struct Apply;

template<unsigned I>
struct Apply<I>
{
	template<typename L, typename F, typename... Args>
	static void apply(L &l, F func, const Statement &, Args &&... args)
	{
		(l.*func)(std::forward<Args>(args)...);
	}

	template<typename L, typename F, typename... Args>
	static void apply(L &l, F func, const ArgumentStore &, Args &&... args)
	{
		(l.*func)(std::forward<Args>(args)...);
	}
};

template<unsigned I, typename Head, typename... Tail>
struct Apply<I, Head, Tail...>
{
	template<typename L, typename F, typename... Args>
	static void apply(L &l, F func, const Statement &st, Args &&... args)
	{
		Apply<I+1, Tail...>::apply(l, func, st, std::forward<Args>(args)..., std::move(st.args[I].get<Head>()));
	}

	template<typename L, typename F, typename... Args>
	static void apply(L &l, F func, const ArgumentStore &as, Args &&... args)
	{
		Apply<I+1, Tail...>::apply(l, func, as, std::forward<Args>(args)..., std::move(as.get<Head>(I)));
	}
};


template<typename L, typename... Args>
class LoaderFuncN: public LoaderAction
{
protected:
	typedef void (L::*FuncType)(Args...);

	FuncType func;

public:
	LoaderFuncN(FuncType f): func(f) { }

	void execute(Loader &l, const Statement &st) const override
	{
		Apply<0, Args...>::apply(dynamic_cast<L &>(l), func, st);
	}

	void execute(Loader &l, const ArgumentStore &as) const override
	{
		Apply<0, Args...>::apply(dynamic_cast<L &>(l), func, as);
	}

	std::string get_signature() const override
	{ return create_signature<Args...>(); }
};


template<typename L, typename B0, typename... Args>
class LoaderFuncNBound1: public LoaderAction
{
protected:
	typedef void (L::*FuncType)(B0, Args...);
	typedef typename std::remove_reference<B0>::type Bound0Type;

	FuncType func;
	Bound0Type bound0;

public:
	LoaderFuncNBound1(FuncType f, const B0 &b0): func(f), bound0(b0) { }
	LoaderFuncNBound1(FuncType f, B0 &&b0): func(f), bound0(std::move(b0)) { }

	void execute(Loader &l, const Statement &st) const override
	{
		Apply<0, Args...>::apply(dynamic_cast<L &>(l), func, st, bound0);
	}

	void execute(Loader &l, const ArgumentStore &as) const override
	{
		Apply<0, Args...>::apply(dynamic_cast<L &>(l), func, as, bound0);
	}

	std::string get_signature() const override
	{ return create_signature<Args...>(); }
};


template<typename L, typename T0>
class LoadValue1: public LoaderAction
{
private:
	typedef T0 L::*Pointer0Type;

	Pointer0Type ptr0;

public:
	LoadValue1(Pointer0Type p0): ptr0(p0) { }

	void execute(Loader &l, const Statement &st) const override
	{
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr0 = st.args[0].get<T0>();
	}

	void execute(Loader &l, const ArgumentStore &as) const override
	{
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr0 = as.get<T0>(0);
	}

	std::string get_signature() const override
	{ return std::string(1, TypeInfo<T0>::signature); }
};


template<typename L, typename T0>
class LoadValue1<L, T0 *>: public LoaderAction
{
private:
	typedef T0 *L::*Pointer0Type;

	Pointer0Type ptr0;

public:
	LoadValue1(Pointer0Type p0): ptr0(p0) { }

	void execute(Loader &l, const Statement &st) const override
	{
		typename L::Loader &ldr = dynamic_cast<typename L::Loader &>(l);
		ldr.get_object().*ptr0 = &ldr.get_collection().template get<T0>(st.args[0].get<std::string>());
	}

	void execute(Loader &l, const ArgumentStore &as) const override
	{
		typename L::Loader &ldr = dynamic_cast<typename L::Loader &>(l);
		ldr.get_object().*ptr0 = &ldr.get_collection().template get<T0>(as.get<std::string>(0));
	}

	std::string get_signature() const override
	{ return std::string(1, TypeInfo<std::string>::signature); }
};


template<typename L, typename T0, typename T1>
class LoadValue2: public LoaderAction
{
private:
	typedef T0 L::*Pointer0Type;
	typedef T1 L::*Pointer1Type;

	Pointer0Type ptr0;
	Pointer1Type ptr1;

public:
	LoadValue2(Pointer0Type p0, Pointer1Type p1): ptr0(p0), ptr1(p1) { }

	void execute(Loader &l, const Statement &st) const override
	{
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr0 = st.args[0].get<T0>();
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr1 = st.args[1].get<T1>();
	}

	void execute(Loader &l, const ArgumentStore &as) const override
	{
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr0 = as.get<T0>(0);
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr1 = as.get<T1>(1);
	}

	std::string get_signature() const override
	{
		std::string result;
		result += TypeInfo<T0>::signature;
		result += TypeInfo<T1>::signature;
		return result;
	}
};

} // namespace DataFile
} // namespace Msp

#endif
