#ifndef MSP_DATAFILE_PACKSOURCE_H_
#define MSP_DATAFILE_PACKSOURCE_H_

#include <msp/core/noncopyable.h>
#include <msp/core/refptr.h>
#include "collectionsource.h"
#include "objectloader.h"

namespace Msp {
namespace DataFile {

/**
A source that loads data from pack files.  As opposed to plain collection
files, pack files are composed from a number of logical files.  They also
contain an index of objects contained in the pack and which logical files they
are in.  This allows the pack to be loaded in a piecewise manner instead of all
at once.

It's possible for a pack file to contain plain collection files as well.  When
an object from such a file is requested, the entire sub-collection it is stored
in is loaded.
*/
class PackSource: public CollectionSource
{
public:
	struct FileInfo
	{
		std::string name;
		IO::SeekOffset size = 0;
	};

private:
	class File;
	class Object;

	typedef std::map<std::string, const File *> FileMap;
	typedef std::map<std::string, const Object *> ObjectMap;

	class Pack
	{
	public:
		class Loader: public ObjectLoader<Pack>
		{
		public:
			Loader(Pack &);
		private:
			void file(const std::string &);
		};

	private:
		std::string filename;
		IO::Seekable *io = nullptr;
		IO::SeekOffset base_offset = 0;
		std::list<File> files;

	public:
		Pack(IO::Seekable *, const std::string &);
		Pack(const Pack &);

		const std::string &get_filename() const { return filename; }
		IO::Seekable *get_io() const { return io; }
		IO::SeekOffset get_base_offset() const { return base_offset; }

		void collect_files(FileMap &, const std::string &) const;
		void translate_files(FileMap &, const Pack &) const;
	};

	class File
	{
	public:
		class Loader: public ObjectLoader<File>
		{
		public:
			Loader(File &);
		private:
			void finish() override;
			void object(const std::string &, const std::string &);
		};

	private:
		const Pack &pack;
		std::string filename;
		IO::SeekOffset offset = 0;
		IO::SeekOffset length = 0;
		bool collection = false;
		std::list<Object> objects;

	public:
		File(const Pack &, const std::string &);
		File(const File &, const Pack &);

		RefPtr<IO::Seekable> open() const;
		const std::string &get_filename() const { return filename; }
		FileInfo get_info() const;
		std::string get_full_name() const;
		bool is_collection() const { return collection; }

		void collect_objects(ObjectMap &) const;
		void translate_objects(ObjectMap &, const File &) const;
	};

	class Object
	{
	private:
		const File &file;
		std::string name;
		std::string keyword;

	public:
		Object(const File &, const std::string &, const std::string &);
		Object(const Object &, const File &);

		const File &get_file() const { return file; }
		const std::string &get_name() const { return name; }
		const std::string &get_keyword() const { return keyword; }
	};

	std::list<Pack> packs;
	FileMap files;
	ObjectMap objects;

public:
	PackSource() = default;
	PackSource(const PackSource &);
	PackSource &operator=(const PackSource &);
private:
	void translate_maps(const PackSource &);

public:
	/// Adds a pack file to load objects from.  The index is read immediately.
	void add_pack_file(const std::string &);

	/** Adds a pack file with a regex to filter logical files.  The index is
	read on the first call; subsequent calls will use cached data. */
	void add_pack_file(const std::string &, const std::string &);

	/** Adds a pack from an existing seekable I/O object.  The same object is
	used for all accesses to the pack, so it must not be deleted before the
	PackSource. */
	void add_pack_io(IO::Seekable &, const std::string & = std::string());

	/** Adds a pack from an I/O object with a regex to filter logical files.
	Multiple filters for the same I/O object can be added with repeated calls. */
	void add_pack_io(IO::Seekable &, const std::string &, const std::string &);

private:
	void add_pack(IO::Seekable *, const std::string &, const std::string &);

public:
	/// Returns information about the files in the pack.
	std::list<FileInfo> list_files() const;

	bool is_loadable(const CollectionItemTypeBase &, const std::string &) const override;
	NameList get_names(const CollectionItemTypeBase &) const override;
	void load(Collection &, const CollectionItemTypeBase &, const std::string &) const override;
	IO::Seekable *open(const std::string &) const override;
};

} // namespace DataFile
} // namespace Msp

#endif
