#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include "except.h"
#include "input.h"
#include "textparser.h"
#include "token.h"

using namespace std;

namespace Msp {
namespace DataFile {

TextParser::TextParser(Input &i, const string &s):
	ParserMode(i, s)
{ }

Statement TextParser::parse()
{
	return parse_statement(nullptr);
}

Statement TextParser::parse_statement(const Token *t)
{
	Statement result;
	unsigned sub = 0;

	while(in)
	{
		Token token;
		if(t)
		{
			token = *t;
			t = nullptr;
		}
		else
			token = parse_token();

		if(result.keyword.empty())
		{
			if(token.str.empty())
				break;
			else if(token.type!=Token::IDENTIFIER)
				throw syntax_error(token.str);
			result.keyword = token.str;
			result.valid = true;
			result.source = src;
			result.line = in.get_line_number();
		}
		else if(sub==1)
		{
			if(token.str=="}")
				sub = 2;
			else
			{
				Statement ss = parse_statement(&token);
				result.sub.push_back(ss);
			}
		}
		else if(sub==2)
		{
			if(token.str!=";")
				throw syntax_error(token.str);
			break;
		}
		else if(token.str=="{")
			sub = 1;
		else if(token.str==";")
			break;
		else if(token.type!=Token::SPECIAL)
			result.append_from_token(token);
		else
			throw syntax_error(token.str);
	}

	return result;
}

Token TextParser::parse_token()
{
	int c = 0;
	int comment = 0;

	// Skip over comments and whitespace
	while(in && comment>=0)
	{
		c = in.get();
		int next = in.peek();

		if(c=='/' && next=='/' && !comment)
			comment = 1;
		else if(c=='/' && next=='*' && !comment)
			comment = 2;
		else if(c=='\n' && comment==1)
			comment = 0;
		else if(c=='*' && next=='/' && comment==2)
			comment = 3;
		else if(comment==3)   // Skip the second character of block comment end
			comment = 0;
		else if(c!=-1 && !isspace(c) && !comment)
			comment = -1;
	}

	if(comment>0)  // EOF while in comment
		throw parse_error(string());
	else if(comment==0)  // Didn't hit any non-whitespace
		return Token(Token::SPECIAL, "");

	enum ParseState
	{
		INIT,
		SIGN,
		FLOATEXPINIT,
		FLOATEXPSIGN,
		STRING,
		STRING_ESCAPE,
		STRING_BASE64,
		ACCEPT,
		ZERO,
		DECIMAL,
		HEXADECIMAL,
		OCTAL,
		FLOAT,
		FLOATEXP,
		STRING_END,
		STRING_BASE64_PAD,
		IDENTIFIER
	};

	static Token::Type token_type[]=
	{
		Token::SPECIAL,
		Token::SPECIAL,
		Token::SPECIAL,
		Token::SPECIAL,
		Token::SPECIAL,
		Token::SPECIAL,
		Token::SPECIAL,
		Token::SPECIAL,
		Token::INTEGER,
		Token::INTEGER,
		Token::INTEGER,
		Token::INTEGER,
		Token::FLOAT,
		Token::FLOAT,
		Token::STRING,
		Token::STRING,
		Token::IDENTIFIER
	};

	ParseState state = INIT;
	string buf;

	while(in || state==INIT)
	{
		if(state!=INIT)
			c = in.get();
		int next = in.peek();

		buf += c;

		switch(state)
		{
		case INIT:
			if(c=='0')
				state = ZERO;
			else if(c=='-' || c=='+')
				state = SIGN;
			else if(c=='.')
				state = FLOAT;
			else if(c=='"')
				state = STRING;
			else if(c=='=')
				state = STRING_BASE64;
			else if(c=='{' || c=='}' || c==';')
				return Token(Token::SPECIAL, string(1, c));
			else if(isdigit(c))
				state = DECIMAL;
			else if(isalpha(c) || c=='_' || c=='\\')
				state = IDENTIFIER;
			else
				throw parse_error(buf);
			break;

		case SIGN:
			if(c=='0')
				state = ZERO;
			else if(isdigit(c))
				state = DECIMAL;
			else if(c=='.')
				state = FLOAT;
			else
				throw parse_error(buf);
			break;

		case ZERO:
			if(c=='x')
				state = HEXADECIMAL;
			else if(isdigit(c))
				state = OCTAL;
			else if(c=='.')
				state = FLOAT;
			else
				throw parse_error(buf);
			break;

		case DECIMAL:
			if(c=='.')
				state = FLOAT;
			else if(c=='e' || c=='E')
				state = FLOATEXPINIT;
			else if(!isdigit(c))
				throw parse_error(buf);
			break;

		case HEXADECIMAL:
			if(!isxdigit(c))
				throw parse_error(buf);
			break;

		case OCTAL:
			if(!isodigit(c))
				throw parse_error(buf);
			break;

		case FLOAT:
			if(c=='e' || c=='E')
				state = FLOATEXPINIT;
			else if(!isdigit(c))
				throw parse_error(buf);
			break;

		case FLOATEXPINIT:
			if(c=='+' || c=='-')
				state = FLOATEXPSIGN;
			else if(isdigit(c))
				state = FLOATEXP;
			else
				throw parse_error(buf);
			break;

		case FLOATEXPSIGN:
			if(isdigit(c))
				state = FLOATEXP;
			else
				throw parse_error(buf);
			break;

		case FLOATEXP:
			if(!isdigit(c))
				throw parse_error(buf);
			break;

		case STRING:
			if(c=='\\')
				state = STRING_ESCAPE;
			else if(c=='"')
				state = STRING_END;
			break;

		case STRING_ESCAPE:
			state = STRING;
			break;

		case STRING_BASE64:
			if(c=='=')
				state = STRING_BASE64_PAD;
			else if(!isalnum(c) && c!='+' && c!='/')
				throw parse_error(buf);
			break;

		case STRING_BASE64_PAD:
			if(c!='=')
				throw parse_error(buf);
			break;

		case IDENTIFIER:
			if(!isalpha(c) && !isdigit(c) && c!='_' && c!='-' && c!='/')
				throw parse_error(buf);
			break;

		case STRING_END:
			throw parse_error(buf);

		default:
			throw logic_error("bad parser state");
		}

		if(is_delimiter(next) && state>=ACCEPT)
		{
			if(state==IDENTIFIER && buf[0]=='\\')
				return Token(Token::IDENTIFIER, buf.substr(1));
			else if(state==STRING_END)
				return Token(Token::STRING, c_unescape(buf.substr(1, buf.size()-2)));
			else if(state==STRING_BASE64 || state==STRING_BASE64_PAD)
				return Token(Token::STRING, base64_decode(buf));
			else
				return Token(token_type[state], buf);
		}
	}

	return Token(Token::SPECIAL, "");
}

bool TextParser::is_delimiter(int c)
{
	return (isspace(c) || c=='{' || c=='}' || c==';' || c=='/');
}

bool TextParser::isodigit(int c)
{
	return (c>='0' && c<='7');
}

string TextParser::base64_decode(const string &data)
{
	string bin;
	bin.reserve(data.size()*3/4);
	unsigned accum = 0;
	unsigned a_bits = 0;
	for(char c: data)
	{
		unsigned d;
		if(c>='A' && c<='Z')
			d = c-'A';
		else if(c>='a' && c<='z')
			d = 26+(c-'a');
		else if(c>='0' && c<='9')
			d = 52+(c-'0');
		else if(c=='+')
			d = 62;
		else if(c=='/')
			d = 63;
		else if(c=='=')
			continue;
		else
			throw invalid_argument("TextParser::base64_decode");

		accum = (accum<<6)|d;
		a_bits += 6;

		if(a_bits>=8)
		{
			bin += (accum>>(a_bits-8))&0xFF;
			a_bits -= 8;
		}
	}

	return bin;
}

} // namespace DataFile
} // namespace Msp
